# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006-2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.
#
# Author: Benjamin Kampmann

"""
Checks for updates on the internet
"""

from elisa.core import common
from elisa.core.utils import defer, misc
from elisa.core.log import Loggable
from elisa.core.components.message import Message

# for request and response parsing
from elisa.plugins.http_client.http_client import ElisaHttpClient
from twisted.web2.stream import BufferedStream
from twisted.internet import reactor, error, task

from distutils.version import LooseVersion


"""
the url used for lookup
"""
HOST = "elisa.fluendo.com"
UPDATES_URL = "http://" + HOST + "/updates/updates/?" \
        "install_date=%(date)s&version=%(version)s&os=%(os)s"\
        "&user_id=%(user_id)s"


class AlreadyRunning(Exception):
    pass


class AvailablePluginUpdatesMessage(Message):

    """
    Message sent when plugins updates are available for download from the
    plugins repository.

    @ivar plugin_dicts: the list of plugins available for update
    @type plugin_dicts: C{list} of {dict}
    """

    def __init__(self, plugin_dicts):
        super(AvailablePluginUpdatesMessage, self).__init__()
        self.plugin_dicts = plugin_dicts


class UpdateChecker(Loggable):

    """
    Helper Class for simple look up of updates on the remote elisa server.
    """

    # check each day, the value is in seconds
    check_interval = 86400

    def __init__(self, install_date, user_id, version):
        super(UpdateChecker, self).__init__()

        # initial data
        self.install_date = install_date
        self.user_id = user_id
        self.version = version
        self.operating_system = misc.get_os_name()

        self.http_client = ElisaHttpClient(HOST, 80)

        self._pending_call = None
        self._current_dfr = None

    def parse_result(self, result):
        """
        Parse the given data into a dictionary. The syntax for result has to be
        a key-value pair per line separated by a colon (':'). Spaces at the
        beginning or the end are stripped from both the key and the value.
        For instance::

            foo: bar
            test: partial
            maybe : maybe not

        @param result: the result to parse
        @type result:  C{str}

        @return:    a dictionary containing the key-value pairs
        @rtype:     C{dict}
        """
        result_dict = {}
        for line in result.splitlines():
            try:
                key, value = line.split(':', 1)
            except ValueError:
                self.warning("Could not split '%s'" % line)
                continue

            result_dict[key.strip()] = value.strip()

        return result_dict

    def request(self):
        """
        Request for update data and parse it
        """
        request_params = {'user_id' : self.user_id,
                          'version' : self.version,
                          'date' : self.install_date,
                          'os' : self.operating_system,}
        request_url = UPDATES_URL % request_params

        dfr = self.http_client.request(request_url)
        dfr.addCallback(self._read_full)
        dfr.addCallback(self.parse_result)
        return dfr

    def start(self, callback):
        """
        Sets up an automatic loop of update url calls starting right now.
        Everytime a result is received the callback is triggered with a
        dictionary of the parsed result as argument. The next iteration is done
        every L{check_interval}-seconds.

        @raises AlreadyRunning: if the method was already called before. It is only
        allowed to call this method once. It is mandatory to call L{stop} before
        calling start again.
        """
        if self._pending_call is not None or self._current_dfr is not None:
            raise AlreadyRunning("Don't call me twice")

        self._auto_request(callback)

    def _got_response(self, result, callback):
        if not result:
            self.warning("response of the Server was empty!")
            return

        dfr = callback(result)
        return dfr

    def _response_failed(self, failure):
        if failure.type != defer.CancelledError:
            self.warning("Problem when requesting for" \
                    "update: %s" % failure)

        # in any case, eat the error
        return None

    def _check_for_plugin_updates(self, result):
        plugin_registry = common.application.plugin_registry

        def get_plugin_updates_list(cache_file):
            plugins = \
                plugin_registry.get_downloadable_plugins(reload_cache=True)
            installed = list(plugin_registry.get_plugin_names())
            updates = []

            def iterate_plugins(plugins, installed, updates):
                for plugin in plugins:
                    name = plugin['name']
                    if name not in installed:
                        continue
                    current = plugin_registry.get_plugin_by_name(name)
                    current_version = LooseVersion(current.version)
                    new_version = LooseVersion(plugin['version'])
                    if new_version <= current_version:
                        continue
                    yield updates.append(plugin)

            def send_message(result, updates):
                if len(updates) > 0:
                    msg = AvailablePluginUpdatesMessage(updates)
                    common.application.bus.send_message(msg)
                    return updates

            updates_dfr = task.coiterate(iterate_plugins(plugins, installed,
                                                         updates))
            updates_dfr.addCallback(send_message, updates)
            return updates_dfr

        def failed_update(failure):
            # Swallow all the errors.
            return None

        dfr = plugin_registry.update_cache()
        dfr.addCallback(get_plugin_updates_list)
        dfr.addErrback(failed_update)
        return dfr

    def _reset_pending_call(self, result, callback):
        pending = reactor.callLater(self.check_interval,
                self._auto_request, callback)
        self._pending_call = pending

    def _auto_request(self, callback):
        self._current_dfr = dfr = self.request()
        dfr.addCallback(self._got_response, callback)
        dfr.addErrback(self._response_failed)
        dfr.addCallback(self._check_for_plugin_updates)
        dfr.addCallback(self._reset_pending_call, callback)

    def stop(self):
        """
        Stop any pending loop or http calls
        """
        try:
            self._pending_call.cancel()
        except (AttributeError, error.AlreadyCalled, error.AlreadyCancelled):
            pass

        try:
            self._current_dfr.cancel()
        except (AttributeError, defer.AlreadyCalledError):
            pass

        self._pending_call = None
        self._current_dfr = None

    def _read_full(self, response):

        buf_stream = BufferedStream(response.stream)
        read_dfr = defer.maybeDeferred(buf_stream.readExactly)

        return read_dfr
